#include "Problem.h"

// output functions
std::ostream& operator<<(std::ostream& os, lecture& l)
{
    (os)<<"["<<l.id<<": "<<l.slot<<","<<l.room<<"]";
    return os;
}

std::ostream& operator<<(std::ostream& os, course& c)
{
    (os)<<c.name<<" (lec.: ";
    for (vector<lecture>::iterator p=c.lectures.begin(); p!=c.lectures.end(); p++)
        (os)<<*p<<" ";
    (os)<<") ";
    (os)<<c.teacher;
    return (os);
}

std::ostream& operator<<(std::ostream& os, teacher& t)
{
    cout<<t.name<<": ";
    for (vector<string>::iterator ps=t.courses.begin(); ps!=t.courses.end(); ps++)
        cout<<*ps<<" ";
    cout<<" Un.: ";
    for (vector<int>::iterator ps=t.unavailabilities.begin(); ps
            !=t.unavailabilities.end(); ps++)
        cout<<*ps<<" ";
    return os;
}

std::ostream& operator<<(std::ostream& os, student& s)
{
    cout<<s.name<<": ";
    for (vector<string>::iterator ps=s.courses.begin(); ps!=s.courses.end(); ps++)
        cout<<*ps<<" ";
    return os;
}

// ----------------------------------------------------------------
int Problem::translate(char day, int time)
{
    int d, t;
    switch (day)
    {
    case 'M':
        d=0;
        break;
    case 'T':
        d=1;
        break;
    case 'W':
        d=2;
        break;
    case 'H':
        d=3;
        break;
    case 'F':
        d=4;
        break;
    default:
        cerr<<"Day not standard"<<endl;
        exit(0);
        break;
    }
    switch (time)
    {
    case 8:
    case 10:
    case 12:
    case 14:
    case 16:
        break;
    default:
        cerr<<time<<" Time not standard"<<endl;
        exit(0);
        break;
    }
    t=(time/2)-4;
    return d*5+t;
}

pair<char, int> Problem::translate_back(int slot)
{
    int d = slot / 5;
    int t = slot-(d*5);
    int hour = (t + 4) * 2;
    char day[5]= { 'M', 'T', 'W', 'H', 'F' };
    return pair<char, int>(day[d], hour);
}


// ----------------------------------------------------------------
void Problem::read_electives(string instance)
{
    ifstream in;
    map<string,course>::iterator pt;
    string fileinstance;
    int lec_id;
    string tmp;
    id_lectures.clear();
    fileinstance = instance+"/electives.dat";
    in.open(fileinstance.c_str());
    if (!in)
    {
        cout<< "Couldn't open the file "<<fileinstance<<endl;
        cout<<"Press return to continue"<<endl;
        char key;
        cin.get(key);
        exit(1);
    }
    lec_id=0;
    in >> tmp;
    while (!in.eof())
    {
        if ((pt=elective_courses.find(tmp))==elective_courses.end())
        {
            int n;
            course tmp_c;
            tmp_c.name=tmp;
            in >> n;
            for (int i=0;i<n;i++)
            {
                lecture tmp_l;
                tmp_l.id = lec_id;
                tmp_l.course=tmp;
                tmp_l.slot=-1;
                tmp_l.room=-1;
                tmp_c.lectures.push_back(tmp_l);
                lec_id++;
            }
            tmp_c.min_sep=MIN_SEP;
            tmp_c.max_sep=MAX_SEP;
            tmp_c.teacher="";
            tmp_c.feas_rooms.clear();
            elective_courses.insert(pair<string, course>(tmp, tmp_c));
            // need to reference back each single lecture
            for (vector<lecture>::iterator pl=elective_courses[tmp].lectures.begin();
                    pl!=elective_courses[tmp].lectures.end(); pl++)
                id_lectures.push_back(&(*pl));
        }
        in >> tmp;
    }
    in.close();
}

void Problem::read_mandatory(string instance)
{
    ifstream in;
    map<string,course>::iterator pt;
    string fileinstance;
    int lec_id;
    string tmp;

    char day;
    int time;
    fileinstance = instance+"/mandatory.dat";
    in.open(fileinstance.c_str());
    if (!in)
    {
        cout<< "Couldn't open the file "<<fileinstance<<endl;
        cout<<"Press return to continue"<<endl;
        char key;
        cin.get(key);
        exit(1);
    }
    lec_id=0;
    in >> tmp;
    while (!in.eof())
    {
        if ((pt=mandatory_courses.find(tmp))==mandatory_courses.end())
        {
            course tmp_c;
            tmp_c.name=tmp;
            tmp_c.teacher="";
            lecture tmp_l;
            tmp_l.id=lec_id;
            tmp_l.course=tmp;
            in >> day >> time;
            tmp_l.slot=translate(day, time);
            tmp_l.room=0;
            tmp_c.lectures.push_back(tmp_l);
            mandatory_courses.insert(pair<string, course>(tmp, tmp_c));
            //mandatory_lecture_course.push_back(tmp);
            lec_id++;
        }
        else
        {
            lecture tmp_l;
            tmp_l.id=lec_id;
            tmp_l.course=tmp;
            in >> day >> time;
            tmp_l.slot=translate(day, time);
            tmp_l.room=DUMMY_FOR_MANDATORY;
            mandatory_courses[tmp].lectures.push_back(tmp_l);
            lec_id++;
        }
        in >> tmp;
    }
    in.close();
}

void Problem::read_teaches(string instance)
{
    ifstream in;

    string fileinstance;
    string tmp;

    fileinstance = instance+"/teaches.dat";
    in.open(fileinstance.c_str());
    if (!in)
    {
        cout<< "Couldn't open the file "<<fileinstance<<endl;
        cout<<"Press return to continue"<<endl;
        char key;
        cin.get(key);
        exit(1);
    }
    id_teachers.clear();
    int id=0;
    in >> tmp;
    string crs;
    map<string, teacher >::iterator pt;
    while (!in.eof())
    {
        if ((pt=teachers.find(tmp))==teachers.end())
        {
            teacher t_tmp;
            t_tmp.name=tmp;
            t_tmp.courses.clear();
            t_tmp.unavailabilities.clear();
            in >> crs;
            t_tmp.courses.push_back(crs);
            t_tmp.id=id;
            teachers.insert(pair<string, teacher>(tmp, t_tmp));
            id_teachers.push_back(tmp);
            id++;
        }
        else
        {
            in >> crs;
            teachers[tmp].courses.push_back(crs);
        }
        if (mandatory_courses.find(crs)!=mandatory_courses.end())
            mandatory_courses[crs].teacher=tmp;
        else if (elective_courses.find(crs)!=elective_courses.end())
            elective_courses[crs].teacher=tmp;
        else
        {
            cerr<<"teaches :"<<instance<<" "<<crs<<endl;
            //exit(0);
        }

        in >> tmp;
    }
    in.close();
}

void Problem::read_teacher_unavailable(string instance)
{
    ifstream in;

    string fileinstance;
    string tmp;

    fileinstance = instance+"/teacher_unavailable.dat";
    in.open(fileinstance.c_str());
    if (!in)
    {
        cout<< "Couldn't open the file "<<fileinstance<<endl;
        cout<<"Press return to continue"<<endl;
        char key;
        cin.get(key);
        exit(1);
    }
    in >> tmp;
    char day;
    int time;
    map<string, teacher >::iterator pt;
    while (!in.eof())
    {
        if ((pt=teachers.find(tmp))!=teachers.end())
        {
            in >> day >> time;
            teachers[tmp].unavailabilities.push_back(translate(day, time));
        }
        else
        {
            in >> day >> time;
            cerr<<"Something wrong in reading the file teacher_unavailable.dat"<<endl;
            exit(0);
        }
        in >> tmp;
    }
    in.close();
}


void Problem::read_takes(string instance)
{
    ifstream in;

    string fileinstance;
    string tmp;

    id_students.clear();
    fileinstance = instance+"/takes.dat";
    in.open(fileinstance.c_str());
    if (!in)
    {
        cout<< "Couldn't open the file "<<fileinstance<<endl;
        cout<<"Press return to continue"<<endl;
        char key;
        cin.get(key);
        exit(1);
    }
    in >> tmp;
    string crs;
    map<string,student >::iterator pt;
    while (!in.eof())
    {
        if ((pt=students.find(tmp))==students.end())
        {
            student s_tmp;
            s_tmp.name=tmp;
            s_tmp.courses.clear();
            in >> crs;
            s_tmp.courses.push_back(crs);
            students.insert(pair<string, student>(tmp, s_tmp));
            id_students.push_back(tmp);
        }
        else
        {
            in >> crs;
            students[tmp].courses.push_back(crs);
        }
        if (mandatory_courses.find(crs)!=mandatory_courses.end())
            mandatory_courses[crs].students.push_back(tmp);
        else if (elective_courses.find(crs)!=elective_courses.end())
            elective_courses[crs].students.push_back(tmp);
        else
        {
            cerr<<"students: "<<instance<<" "<<crs<<endl;
            //exit(0);
        }
        in >> tmp;
    }
    in.close();
}


void Problem::read_rooms(string instance)
{
    ifstream in;

    string fileinstance;
    string tmp;
    id_rooms.clear();
    fileinstance = instance+"/rooms.dat";
    in.open(fileinstance.c_str());
    if (!in)
    {
        cout<< "Couldn't open the file "<<fileinstance<<endl;
        cout<<"Press return to continue"<<endl;
        char key;
        cin.get(key);
        exit(1);
    }
    in >> tmp;
    map<string, room >::iterator pt;
    int id=0;
    while (!in.eof())
    {
        if ((pt=rooms.find(tmp))==rooms.end())
        {
            room r_tmp;
            r_tmp.name=tmp;
            r_tmp.id =id++;
            r_tmp.unavailabilities.clear();
            rooms.insert(pair<string, room> (tmp,r_tmp));
            id_rooms.push_back(tmp);
        }
        in >> tmp;
    }
    in.close();
/*
    bool at_least_one=false;
    ofstream os("course_rooms.dat");
    map<string,course>::iterator p;
    for (p=elective_courses.begin();p!=elective_courses.end();p++)
    {
        at_least_one=false;
        for (pt=rooms.begin();pt!=rooms.end();pt++)
        {
            if ((double) rand()/RAND_MAX <0.5)
            {
                at_least_one=true;
                os<<p->second.name<<"\t"<<pt->second.name<<endl;
            }
        }
        if (!at_least_one)
            os<<p->second.name<<"\t"<<rooms.begin()->second.name<<endl;
    }
    os.close();
    */
}

void Problem::read_room_unavailable(string instance)
{
    ifstream in;

    string fileinstance;
    string tmp;

    fileinstance = instance+"/room_unavailable.dat";
    in.open(fileinstance.c_str());
    if (!in)
    {
        cout<< "Couldn't open the file "<<fileinstance<<endl;
        cout<<"Press return to continue"<<endl;
        char key;
        cin.get(key);
        exit(1);
    }
    in >> tmp;
    char day;
    int time;
    map<string, room >::iterator pt;
    while (!in.eof())
    {
        if ((pt=rooms.find(tmp))!=rooms.end())
        {
            in >> day >> time;
            rooms[tmp].unavailabilities.push_back(translate(day, time));
        }
        else
        {
            in >> day >> time;
            cerr<<"Something wrong in reading the file room_unavailable.dat"<<endl;
            exit(0);
        }
        in >> tmp;
    }
    in.close();
}

void Problem::read_course_rooms(string instance)
{
    ifstream in;

    string fileinstance;
    string c_tmp, r_tmp ;

    fileinstance = instance+"/course_rooms.dat";
    in.open(fileinstance.c_str());
    if (!in)
    {
        cout<< "Couldn't open the file "<<fileinstance<<endl;
        //cout<<"Press return to continue"<<endl;
        //char key;
        //cin.get(key);
        //exit(1);
        //in.close();
        return;
    }
    in >> c_tmp >> r_tmp;
    map<string, room >::iterator pt;
    while (!in.eof())
    {
        if ((pt=rooms.find(r_tmp))!=rooms.end())
        {
            elective_courses[c_tmp].feas_rooms.push_back(r_tmp);
        }
        else
        {
            cerr<<"Something wrong in reading the file course_rooms.dat"<<endl;
            exit(0);
        }
        in >> c_tmp >> r_tmp;
    }
    in.close();
}



void Problem::read_instance(string instance)
{
    read_electives(instance);
    read_mandatory(instance);
    read_teaches(instance);
    read_teacher_unavailable(instance);
    read_takes(instance);
    read_rooms(instance);
    read_room_unavailable(instance);
    read_course_rooms(instance);

    cout<<"Elective courses:"<<endl;
    for (int i=0; i < (int) id_lectures.size(); i++)
        cout<<elective_courses[id_lectures[i]->course]<<endl;

    cout<<"Mandatory courses:"<<endl;
    for (map<string,course>::iterator pt=mandatory_courses.begin(); pt
            !=mandatory_courses.end(); pt++)
        cout<<(*pt).second<<endl;

    cout<<"Teachers:"<<endl;
    for (map<string,teacher >::iterator p=teachers.begin(); p!=teachers.end(); p++)
        cout<<(*p).second<<endl;

    cout<<"Students:"<<endl;
    for (map<string,student>::iterator p=students.begin(); p!=students.end(); p++)
        cout<<(*p).second<<endl;

}


void Problem::aux_set_conflicts_lectures_of_courses(string c1, string c2)
{
    int l1, l2;
    for (vector<lecture>::iterator pl1=elective_courses[c1].lectures.begin();
            pl1!=elective_courses[c1].lectures.end(); pl1++)
        for (vector<lecture>::iterator pl2=elective_courses[c2].lectures.begin();
                pl2!=elective_courses[c2].lectures.end(); pl2++)
        {
            l1=pl1->id;
            l2=pl2->id;
            assert(l1>=0 && l1 < n_lcts);
            assert(l2>=0 && l2 < n_lcts);
            lecture_conflicts[l1][l2]=1;
            lecture_conflicts[l2][l1]=1;
        }
}


// creation of internal representation
void Problem::process_data()
{
    n_lcts=id_lectures.size();
    lecture_conflicts = new unsigned char*[n_lcts];
    for (int i=0; i<n_lcts; i++)
    {
        lecture_conflicts[i] = new unsigned char [n_lcts];
        memset(lecture_conflicts[i], 0, sizeof(unsigned char)*n_lcts);
    }


    n_stds=students.size();
    student_availability = new unsigned char*[N_TIMESLOTS];
    for (int i=0; i<N_TIMESLOTS; i++)
    {
        student_availability[i] = new unsigned char [n_stds];
        memset(student_availability[i], 0, sizeof(unsigned char)*n_stds);
    }


    n_tchs=teachers.size();
    teacher_availability = new unsigned char*[N_TIMESLOTS];
    for (int i=0; i<N_TIMESLOTS; i++)
    {
        teacher_availability[i] = new unsigned char [n_tchs];
        memset(teacher_availability[i], 0, sizeof(unsigned char)*n_tchs);
    }

    n_rms=id_rooms.size();
    room_availability = new unsigned char*[N_TIMESLOTS];
    for (int i=0; i<N_TIMESLOTS; i++)
    {
        room_availability[i] = new unsigned char [n_rms];
        memset(room_availability[i], 0, sizeof(unsigned char)*n_rms);
    }

    room_feasibility = new unsigned char*[n_lcts];
    for (int i=0; i<n_lcts; i++)
    {
        room_feasibility[i] = new unsigned char [n_rms];
        memset(room_feasibility[i], 0, sizeof(unsigned char)*n_rms);
    }

    // set lecture conflicts due to students
    // set unavailabilities for studnets when they attend mandatory courses
    for (int i=0; i<n_stds; i++)
    {
        student *pt=&students[id_students[i]];
        for (vector<string>::iterator pc1=pt->
                                          courses.begin();
                pc1
                !=pt->courses.end();
                pc1++)
            if (elective_courses.find(*pc1)
                    !=elective_courses.end())
            {
                for (vector<string>::iterator pc2=pc1; pc2!=pt->
                        courses.end();
                        pc2++)
                    if (pc1!=pc2 && elective_courses.find(*pc2)
                            !=elective_courses.end())
                        aux_set_conflicts_lectures_of_courses(*pc1,*pc2);

            }
            else if (mandatory_courses.find(*pc1)
                     !=mandatory_courses.end())
            {
                for (vector<lecture>::iterator pl=mandatory_courses[*pc1].lectures.begin();
                        pl!=mandatory_courses[*pc1].lectures.end(); pl++)
                {
                    assert(pl->slot<N_TIMESLOTS);
                    if (student_availability[pl->slot][i]==0)
                        student_availability[pl->slot][i]=1;
                    else
                    {
                        cerr<<"Mandatory courses overlapping "<<pl->course<<" student "<<students[id_students[i]].name;
                        cerr<<translate_back(pl->slot).first<<" "<<translate_back(pl->slot).second<<endl;
                        //exit(0);
                    }
                }
            }
    }

    // set lecture conflicts for teachers who share courses
    // set teachers unavailabilities for teachers when they mandatory courses.
    for (int t=0; t<n_tchs; t++)
    {
        teacher *pt=&teachers[id_teachers[t]];

        for (vector<int>::iterator p=pt->
                                     unavailabilities.begin();
                p!=pt->unavailabilities.end();
                p++)
        {
            teacher_availability[(*p)][t]=1;
        }
        for (vector<string>
                ::iterator pc1=pt->
                               courses.begin();
                pc1
                !=pt->courses.end();
                pc1++)
            if (elective_courses.find(*pc1)
                    !=elective_courses.end())
            {
                for (vector<string>::iterator pc2=pc1; pc2!=pt->
                        courses.end();
                        pc2++)
                    if (pc1!=pc2 && elective_courses.find(*pc2)
                            !=elective_courses.end())
                        aux_set_conflicts_lectures_of_courses(*pc1,*pc2);

            }
            else if (mandatory_courses.find(*pc1)
                     !=mandatory_courses.end())
            {
                for (vector<lecture>::iterator pl=mandatory_courses[*pc1].lectures.begin();
                        pl!=mandatory_courses[*pc1].lectures.end(); pl++)
                {
                    //cerr<<"here"<<endl;
                    assert(pl->slot<N_TIMESLOTS);
                    pt->unavailabilities.push_back(pl->slot);
                    teacher_availability[pl->slot][t]=1;
                }
            }
    }

    // output
    cout<<"Matrix of student unavailability [N_TIMESLOTS x n_stds]"<<endl;
    for (int i=0; i<N_TIMESLOTS; i++)
    {
        for (int j=0; j<n_stds; j++)
            printf("%d ", student_availability[i][j]);
        printf("\n");
    }

    cout<<"Matrix of teacher unavailability [N_TIMESLOTS x n_tchs]"<<endl;
    for (int i=0; i<N_TIMESLOTS; i++)
    {
        for (int j=0; j<n_tchs; j++)
            printf("%d ", teacher_availability[i][j]);
        printf("\n");
    }

    cout<<"Matrix of lecture conflicts [n_lcts x n_lcts]"<<endl;
    for (int i=0; i<n_lcts; i++)
    {
        for (int j=0; j<n_lcts; j++)
            printf("%d ", lecture_conflicts[i][j]);
        printf("\n");
    }

    // create data student_x_lecture
    students_x_lecture.clear();
    students_x_lecture.resize(n_lcts);
    for (int s=0; s< n_stds; s++)
    {
        for (vector<string>::iterator pt=
                    students[id_students[s]].courses.begin(); pt
                !=students[id_students[s]].courses.end(); pt++)
            if (elective_courses.find(*pt)!=elective_courses.end())
                for (vector<lecture>::iterator pl=
                            elective_courses[*pt].lectures.begin(); pl
                        !=elective_courses[*pt].lectures.end(); pl++)
                    students_x_lecture[pl->id].push_back(s);
    }
    cout<<"Student attendace at the mandatory courses"<<endl;
    for (map<string, course>::iterator p=mandatory_courses.begin(); p
            !=mandatory_courses.end(); p++)
        cout<<p->first<<": " <<p->second.students.size()<<endl;
    cout<<"Student attendace at the elective courses (2 lectures per course)"
    <<endl;
    for (int l=0; l<n_lcts; l++)
        cout<<l<<" "<<*(id_lectures[l]) <<": " <<students_x_lecture[l].size()
        <<endl;

    // set time slots that are unavailable for rooms
    for (int r=0; r<n_rms; r++)
    {
        room *pt=&rooms[id_rooms[r]];
        for (vector<int>::iterator p=pt->
                                     unavailabilities.begin();
                p!=pt->unavailabilities.end();
                p++)
        {
            room_availability[(*p)][r]=1;
        }
    }

    for (map<string,course>::iterator pc=elective_courses.begin();
            pc!=elective_courses.end();pc++)
        for (vector<lecture>::iterator pl=pc->
                                          second.lectures.begin();
                pl!=pc->second.lectures.end();
                pl++)
            for (vector<string>
                    ::iterator pr=pc->second.feas_rooms.begin();
                    pr!=pc->second.feas_rooms.end();
                    pr++)
                room_feasibility[pl->id][rooms[*pr].id]=1;

    cout<<"Matrix of feasible rooms per lecture [n_lcts x n_rms]"<<endl;
    for (int i=0; i<n_lcts; i++)
    {
        for (int j=0; j<n_rms; j++)
            printf("%d ", room_feasibility[i][j]);
        printf("\n");
    }

}
