#ifndef PROBLEM_H_
#define PROBLEM_H_

#include "Globals.h"

struct lecture
{
    int id;
    string course;
    int slot;
    int room;
};

struct course
{
    int id;
    string name;
    vector<lecture> lectures;
    string teacher;
    vector<string> students;
    int min_sep;
    int max_sep;
    vector<string> feas_rooms;
};

struct teacher
{
    int id;
    string name;
    vector<string> courses;
    vector<int> unavailabilities;
};

struct student
{
    int id;
    string name;
    vector<string> courses;
};

struct room
{
    int id;
    string name;
    vector<int> unavailabilities;
};


class Problem
{
public:
    Problem()
    {}

    ~Problem()
    {}

    // ----------------------------------------------------------------
    map<string, course > elective_courses;
    map<string, course > mandatory_courses;
    map<string, student > students;
    map<string, teacher > teachers;
    map<string, room> rooms;
    // ----------------------------------------------------------------
    // for internal representation
    // internally names are mapped to integers identifiers
    // other data structures beside these might be useful
    int n_lcts;
    int n_stds;
    int n_tchs;
    int n_rms;
    vector< lecture* > id_lectures; // elective lectures only
    vector< string > id_students;
    vector< string > id_teachers;
    vector< string > id_rooms;

    unsigned char **lecture_conflicts; // 0 not conflicting; 1 conflicting; determined by shared students and teachers
    unsigned char **student_availability; //0 available; 1 unavailable;
    unsigned char **teacher_availability; //0 available; 1 unavailable;
    unsigned char **room_availability; //0 available; 1 unavailable;
    unsigned char **room_feasibility; //0 unfeasible; 1 feasible;
    vector< vector<int> > students_x_lecture;
    int state;
    int distance_feasibility, cost;
    // ----------------------------------------------------------------
    void read_instance(string instance);
    void process_data();
private:
    int translate(char day, int time);
    pair<char, int> translate_back(int slot);
    void read_electives(string instance);
    void read_mandatory(string instance);
    void read_teaches(string instance);
    void read_teacher_unavailable(string instance);
    void read_takes(string instance);
    void read_rooms(string instance);
     void read_course_rooms(string instance);
    void read_room_unavailable(string instance);
    void aux_set_conflicts_lectures_of_courses(string c1, string c2);
};

class Solution
{
public:
    Solution(Problem *pt): p(pt) {}

    ~Solution()
    {}

    // ----------------------------------------------------------------
    // solution
    vector< pair<int, int > > tt_electives;
    int state;
    int distance_feasibility;
    int cost;


    void read_sln(string instance);
    void random_sln();
    void check_state_sln(bool);
    void evaluate();
private:
    Problem *p;
};

#endif /*PROBLEM_H_*/
