#include "Globals.h"
#include "Problem.h"

//------------------------------------------------------------------
void Solution::read_sln(string instance)
{
    ifstream in;
    tt_electives.resize(p->n_lcts);

    for (int l=0;l<p->n_lcts;l++)
    {
        tt_electives[l].first=-1;
        tt_electives[l].second=-1;
    }
    string tmp;
    in.open(instance.c_str());
    if (!in)
    {
        cout<< "Couldn't open the file: "<<instance<<endl;
        cout<<"Press return to continue"<<endl;
        char key;
        cin.get(key);
        exit(1);
    }
    in >> tmp;

    while (!in.eof())
    {
        if (p->elective_courses.find(tmp)!=p->elective_courses.end())
        {
            vector<lecture>::iterator pl=p->elective_courses[tmp].lectures.begin();
            while (pl!=p->elective_courses[tmp].lectures.end()
                    && pl->slot >= 0)
                pl++;
            if (pl==p->elective_courses[tmp].lectures.end())
            {
                cerr
                <<"Some course has been assigned more lectures than required"
                <<endl;
                exit(0);
            }
            else
            {
                int slot, room;
                in >> slot >> room;
                pl->slot=slot;
                pl->room=room;
                tt_electives[pl->id]=pair<int,int> (pl->slot,pl->room);
            }
        }
        else
        {
            cerr<<"The course "<<tmp
            <<" in the solution file is not an elective course"<<endl;
            exit(0);
        }
        in >> tmp;
    }
    in.close();
}

void Solution::random_sln()
{
    ofstream out("rnd.sln");
    srand(time(NULL));
    for (int l=0; l<p->n_lcts; l++)
    {
        out<<p->id_lectures[l]->course<<" ";
        out<<rand() % N_TIMESLOTS<<" ";
        out<<rand() % p->n_rms<<endl;
    }
    out.close();
}

//------------------------------------------------------------------
void Solution::check_state_sln(bool check_room)
{
    //check if feasible
    distance_feasibility = 0;
    for (int l = 0 ; l < p->n_lcts; l++)
    {
        if (tt_electives[l].second < 0 || tt_electives[l].second >= p->n_rms)
        {
            cout<<" Solution NOT FEASIBLE. A lecture is not scheduled in the seminar room."<<endl;
            state = 1;
            distance_feasibility += p->students_x_lecture[l].size();
        }
    }
    //check if valid
    for (int l = 0 ; l < p->n_lcts; l++)
    {
        int slot = tt_electives[l].first;
        if (slot<0 || slot>=N_TIMESLOTS)
        {
            cout<<"Solution NOT VALID. A lecture has not been assigned to any valid time slot."<<endl;
            state=0;
            continue;
        }
        for (int j = l+1 ; j < p->n_lcts; j++)
            if (tt_electives[j].first==slot && tt_electives[j].second==tt_electives[l].second)
            {
                cout<<"Solution NOT VALID. Two courses sheduled in the same timeslot in the same room."<<endl;
                state=0;
            }

        if (tt_electives[l].second < p->n_rms && p->room_availability[slot][tt_electives[l].second]>0)
        {
            cout<<"Solution NOT VALID. Lecture scheduled in a room not available at the timeslot."<<endl;
            state=0;
        }
        for (vector<int>::iterator ps = p->students_x_lecture[l].begin();ps != p->students_x_lecture[l].end();ps++)
            if (p->student_availability[slot][*ps]>0)
            {
                cout<<"Solution NOT VALID. Student has to attend two lectures of mandatory and elective courses at the same time."<<endl;
                state=0;
            }
        string teacher = p->elective_courses[p->id_lectures[l]->course].teacher;
        if (p->teacher_availability[slot][p->teachers[teacher].id]>0)
        {
            cout<<"Solution NOT VALID. Lecture scheduled in an unavailable slot for a teacher."<<endl;
            state=0;
        }
        for (int j = l+1 ; j < p->n_lcts; j++)
            if (p->lecture_conflicts[l][j] && tt_electives[j].first==slot)
            {
                cout<<"Solution NOT VALID. Student has to attend two lectures of elective courses at the same time."<<endl;
                state=0;
            }
        if (check_room && p->room_feasibility[l][tt_electives[l].second]==0)
        {
            cout<<"Solution NOT FEASIBLE. A lecture is scheduled in an unfeasible room."<<endl;
            state = 0;
        }
    }
}


void Solution::evaluate()
{
    int slot;
    int viol1,viol2,viol3;
    unsigned char **student_schedule;
    student_schedule = new unsigned char*[N_TIMESLOTS];
    for (int i=0; i<N_TIMESLOTS; i++)
    {
        student_schedule[i] = new unsigned char [p->n_stds];
        memset(student_schedule[i], 0, sizeof(unsigned char)*p->n_stds);
    }

    cost=0;
    //------------------------------------------------------------------
    viol1=0;
    for (int l=0; l<p->n_lcts; l++)
    {
        slot = tt_electives[l].first;
        if (slot%5==0 || slot%5==4) //first or last slot of the day
            cost+=p->students_x_lecture[l].size();
    }
    viol1=cost;
    cerr<<"Viol. due to last slot of the day "<<viol1<<endl;
    //------------------------------------------------------------------
    viol2=0;
    for (int l=0; l<p->n_lcts; l++)
        for (vector<int>::iterator ps=p->students_x_lecture[l].begin(); ps!=p->students_x_lecture[l].end();ps++)
            student_schedule[tt_electives[l].first][*ps]=1;

    int count;
    for (int s=0; s<p->n_stds; s++)
    {
        for (int i=0;i<5; i++)
        {
            count=0;
            for (int j=0;j<5; j++)
            {
                slot=5*i+j;
                if (student_schedule[slot][s]>0 || p->student_availability[slot][s]>0)  // count both elective and mandatory courses!
                    count++;
            }
            if (count>3) // || count == 1 
                cost++; // a student has more than three lecutres in day
        }
    }
    viol2=cost-viol1;
    cerr<<"Viol. due to num. of lectures in a day "<<viol2<<endl;    
    //------------------------------------------------------------------
    viol3=0;
    int sep;
    for (map<string, course>::iterator pc=p->elective_courses.begin();pc!=p->elective_courses.end(); pc++)
    {
        vector<lecture>::iterator pl=pc->second.lectures.begin();
        vector<lecture>::iterator plnext=pl;
        plnext++;
        while (plnext != pc->second.lectures.end())
        {
            sep = abs((tt_electives[plnext->id].first / 5)-(tt_electives[pl->id].first / 5));
            if (sep < pc->second.min_sep || sep > pc->second.max_sep)
                cost += pc->second.students.size();
            plnext++;
            pl++;
        }
    }
      viol3=cost-viol1-viol2;
    cerr<<"Viol. due to separation distance "<<viol3<<endl;    
}


